package connectionRules;
import java.util.*;

import utilities.Individual;
import utilities.ModelError;

import learningRules.LearningRule;

/**
* A strategy learning rule that is like reinforcement dynamics.  This learning rule reinforces
* both a players strategy with their payoff but also the neighbors' strategy with the neighbor's
* payoff (in the player's weights).  This can be done at some discount.
*/
public class NeighborReinforce implements LearningRule {
        
        // The neighbor discount value
        private double neighbordiscount;
        /**
        * Sets the neighbor discount rate.  This is the rate at which weights are discounted if
        * the weights come from a neighbors actions.
        * @param n The discount rate (1.0 = no discount)
        */
        public void setNeighDiscount(double n) { neighbordiscount = n; }
        
        /**
        * Returns the neighbor discount rate.
        * @return The discount rate
        */
        public double getNeighDiscount() { return neighbordiscount; }
        
        /**
        * Prepares the individual for a new round of play.  This just aplies the past discount
        * factor to the wieghts.
        * @param i The calling individual
        */
        public void newGeneration(Individual i) throws ModelError {
                for(int j=0;j<i.getWeights().length;j++) 
                        i.setStratWeights(j, (double) i.getWeights(j)*i.getDiscount());
        }
        
        /**
        * This function is run after the payoffs are updated for all individuals.  This adds
        * the weights for both the individual and neighbors strategy (possibly at some discount)
        * @param i The calling invidual
        * @throws ModelError If there is a problem
        */
        public void processPayoff(Individual i) throws ModelError {
                i.addStratWeight(i.getStrat(), i.getPayoff());
                Iterator j = i.getConnected().iterator();
                while(j.hasNext()) {
                        Individual n = (Individual) j.next();
                        i.addStratWeight(n.getStrat(), (double) n.getPayoff()*neighbordiscount);
                }
        }
        
        //pickStrat():  Chosses a strategy for individual i, based on the current wieghts
        /**
        * Choose a strategy for the individual based on the wieghts by the standard reinforcement
        * method.
        * @param i The calling Individual
        * @return The strategy choosen
        */
        public int pickStrat(Individual i) {
                double total = 0.0;
                double randnum;
                int strat=-1;
                
                for(int j=0;j<i.getWeights().length;j++)
                        total += i.getWeights(j);
                randnum = i.getRandom().nextDouble();
                for(int j=0;j<i.getWeights().length;j++) {
                        double subtotal = 0.0;
                        for(int k=0;k<=j;k++)
                                subtotal += i.getWeights(k);
                        if(randnum <= ((double) subtotal / total)) {
                                strat = j;
                                break;
                        }
                }
                return strat;
        }

        /**
        * Returns a NeighborReinforce object. The default discount factor is no discount at all
        */
        public NeighborReinforce() { neighbordiscount = 1.0; }
        
        /**
        * Returns a NeighborReinforce object with specified neigbor discount factor.
        * @param n The neighbor discount factor
        */
        public NeighborReinforce(double n) { neighbordiscount = n; }
}
