package learningRules;

import java.util.HashSet;
import java.util.Iterator;

import utilities.Game;
import utilities.Individual;
import utilities.ModelError;

/**
 * This class implements the Hybrid learning rule.  Every generation each individual
 * averages their neighbors beliefs with their own (giving equal weight to each persons
 * belief) and then adopts it.  Technically, the new beliefs are placed in domain 1 (where
 * the averaged beliefs are in domain 0).  This requires the calling model to copy
 * beliefs from domain 1 to domain 0 after calling newGeneration();
 * 
 * @author Kevin Zollman
 *
 */
public class HybridLearning implements LearningRule {

	Game games[];
	
	/**
	 * The new generation method implements the averaging.
	 */
	public void newGeneration(Individual i) throws ModelError {
		HashSet neigh = i.getConnected();
        double[][] neighborbeliefs = new double[neigh.size()+1][games.length];
		int count = 0;
		
		//get my beliefs
		for(int k=0;k<games.length;k++) 
			neighborbeliefs[count][k] = i.getBelief(0,k);
		count++;
		//get my neighbors beliefs
		synchronized(neigh) {
        Iterator j = neigh.iterator();
        while(j.hasNext()) {
        	Individual n = (Individual) j.next();
        	for(int k=0;k<games.length;k++) 
    			neighborbeliefs[count][k] = n.getBelief(0,k);
    		count++;
        }
        }
		
		double[] newbeliefs = new double[games.length];
		
		for(int k=0;k<games.length;k++) {
			newbeliefs[k] = 0.0;
			for(int m=0;m<neigh.size()+1;m++) {
				newbeliefs[k] += (double) ((double) 1/(neigh.size()+1))*neighborbeliefs[m][k];
			}
		}
		
		i.setBeliefs(newbeliefs, 1);
	}

	
	/**
	 * Chooses the strategy with the highest expectation
	 */
	public int pickStrat(Individual i) {
        double expectation[] = new double[games[0].getStrats()];
        
        for(int j=0;j<games[0].getStrats();j++)
                expectation[j] = 0.0;
        for(int j=0;j<games[0].getStrats();j++) {
                for(int k=0;k<games.length;k++) 
                       expectation[j] += (double) i.getBelief(k)*games[k].getExpectedPayoff(j);
        }
        
        double highestexpectation = 0.0;
        int strat = 0;
        for(int j=0;j<expectation.length;j++) {
                if (expectation[j] > highestexpectation) {
                        strat = j;
                        highestexpectation = expectation[j];
                }
        }
        
        return strat;
	}

	/**
	 * Updates beliefs based on the payoffs of neighbors.  Payoff == -1 means
	 * no experiment was performed and their payoff is ignored.
	 */
	public void processPayoff(Individual i) throws ModelError {
		  double denom = 0.0;
          double[] newBeliefs = new double[games.length];
          
          for(int k=0;k<games.length;k++) 
        	  newBeliefs[k] = i.getBelief(k);
          

          
          // first based on your payoff
          if(i.getPayoff() != -1) {
          for(int k=0;k<games.length;k++) 
                  denom += (double) games[k].getProb(i.getStrat(), i.getPayoff())*i.getBelief(k);
          
          for(int k=0;k<games.length;k++) {
                  newBeliefs[k] = (double) ( (double) games[k].getProb(i.getStrat(), i.getPayoff()) * i.getBelief(k))/denom;
                  if(Double.isNaN(newBeliefs[k]))
                          throw new ModelError("NaN Belief: " + games[k].getProb(i.getStrat(), i.getPayoff()) +"*" + i.getBelief(k) + " / " + denom);
          }
          }
          
          //then your neighbors
          HashSet neigh = i.getConnected();
          synchronized(neigh) {
          Iterator j = neigh.iterator();
          while(j.hasNext()) {
                  Individual n = (Individual) j.next();
                  
                  if(n.getPayoff() != -1) {
                  denom = 0.0;
                  for(int k=0;k<games.length;k++) 
                          denom += (double) games[k].getProb(n.getStrat(), n.getPayoff())*newBeliefs[k];
                  for(int k=0;k<games.length;k++) {
                          newBeliefs[k] = (double) ((double) games[k].getProb(n.getStrat(), n.getPayoff()) * newBeliefs[k])/denom;
                          if(Double.isNaN(newBeliefs[k]))
                                  throw new ModelError("NaN Belief (" +i.getId() + "/" + n.getId() + "): " + games[k].getProb(n.getStrat(), n.getPayoff()) +"*" + newBeliefs[k] +" / " +denom);
                  }
                  }	
          }
          }
          
          i.setBeliefs(newBeliefs);
	}
	
	public HybridLearning(Game g[]) { games = g; } 

}
