package learningRules;

import java.util.*;

import utilities.Game;
import utilities.Individual;
import utilities.ModelError;

/**
* A learning rule where individuals learn the payoffs in a n-strategy game.  This learning
*   rule treats the payoffs for each strategy as stable (i.e. non-strategic).  It using beta 
*   distribution learning.
* @Author Kevin Zollman
*/

public class PayoffLearning implements LearningRule {
        
        Game game;
        
        /**
        * This does nothing since there is no discounting or anything.
        * @param i The individual calling the function
        * @throws ModelError (although, since this does nothing it will never be thrown)
        */
        public void newGeneration(Individual i) throws ModelError { }
        
        /**
        * This processes the payoff by updating the agents beliefs about each payoff based 
        *   on the information received this round. 
        * @param i The individual whose beliefs should be updated.
        */
        
        public void processPayoff(Individual i) throws ModelError {
                
                i.addAlpha(i.getPayoff(), i.getStrat());
                i.addBeta(game.getHighestPayoff() - i.getPayoff(), i.getStrat());
                
                HashSet neigh = i.getConnected();
                synchronized(neigh) {
                Iterator j = neigh.iterator();
                while(j.hasNext()) {
                        Individual n = (Individual) j.next();
                        i.addAlpha(n.getPayoff(), n.getStrat());
                        i.addBeta(game.getHighestPayoff() - n.getPayoff(), n.getStrat());
                }
                }
        }
        
        /**
        * Chooses the strategy with the highest expected payoff.  In case of a tie
        *    it chooses the higher numerical strategy.
        * @param i the individual whose strategy is choosen
        * @return the new strategy
        */
        public int pickStrat(Individual i) {
                double[] expectation = new double[game.getStrats()];
                int higheststrat = 0;
                
                for(int j=0;j<expectation.length;j++) {
                        expectation[j] = (double) i.getAlpha(j)/((double) i.getAlpha(j) + i.getBeta(j));
                        if(expectation[j] >= expectation[higheststrat])
                                higheststrat = j;
                }
                return higheststrat;
        }

        public PayoffLearning(Game g) throws ModelError{ game = g; }
}
