package learningRules;

import java.util.*;

import utilities.Game;
import utilities.Individual;
import utilities.ModelError;

/**
* A learning rule where individuals learn the payoffs in a n-strategy game.  This learning
*   rule treats the payoffs for each strategy as stable (i.e. non-strategic).  It using beta 
*   distribution learning.
* @Author Kevin Zollman
*/

public class SmoothPayoffLearning implements LearningRule {
        
        Game game;
        double gamma;
        
        /**
        * This calculates 1/(e^x[0]+e^x[1]+...+e^x[n])
        * It returns a double which is a rounded version of the result
        * @param x The list of exponents
        * @return 1/(e^x[0]+e^x[1]+...+e^x[n])
        */
        static final double probs(double[] x) {
                double total = 0.0;
                for(int i=0;i<x.length;i++) {
                        double subt = Math.exp(x[i]);
                        total += subt;
                }
                return (double) 1/total;
        }
        
        /**
        * This does nothing since there is no discounting or anything.
        * @param i The individual calling the function
        * @throws ModelError (although, since this does nothing it will never be thrown)
        */
        public void newGeneration(Individual i) throws ModelError { }
        
        /**
        * This processes the payoff by updating the agents beliefs about each payoff based 
        *   on the information received this round. 
        * @param i The individual whose beliefs should be updated.
        */
        
        public void processPayoff(Individual i) throws ModelError {
                
                i.addAlpha(i.getPayoff(), i.getStrat());
                i.addBeta(game.getHighestPayoff() - i.getPayoff(), i.getStrat());
                
                HashSet neigh = i.getConnected();
                synchronized(neigh) {
                Iterator j = neigh.iterator();
                while(j.hasNext()) {
                        Individual n = (Individual) j.next();
                        i.addAlpha(n.getPayoff(), n.getStrat());
                        i.addBeta(game.getHighestPayoff() - n.getPayoff(), n.getStrat());
                }
                }
        }
        
        /**
        * This function chooses a strategy for the player.  It uses a probabilistic (smoothed) 
        * best response to its beliefs function.  The function used is an exponentiation function
        * for two strategies the probability of strategy one is 
        * <pre>e^(E(1)/gamma) / *(e^(E(1)/gamma) + e^(E(2)/gamma))</pre> where E(x) is the 
        * expectation of action x and gamma is a parameter which corresponds to the amount
        * of "smoothing" that occurs (i.e. the higher the gamma the further from the 
        * best response correspondence the function is).
        * @param i the individual whose strategy is choosen
        * @return the new strategy
        */
        public int pickStrat(Individual i) {
                double expectation[] = new double[game.getStrats()];
                
                //Calculate the expectation/gamma for each action
                for(int j=0;j<expectation.length;j++)
                        expectation[j] = (double) game.getHighestPayoff() * ((double) i.getAlpha(j)/((double) ((double) i.getAlpha(j) + i.getBeta(j)) * gamma));
                
                
                // Assign probabilities to each action using the exponentiation best response
                //    rule
                // This code looks strange, but this is the most optmized I could come up with
                
                double[] probabilities = new double[expectation.length];
                
                probabilities[expectation.length-1] = 1.0;
                double[] diffs = new double[expectation.length];
                for(int j=0;j<(expectation.length -1);j++) {
                        for(int k=0;k<expectation.length;k++) {
                                diffs[k] = (double) expectation[k] - expectation[j];
                        }
                        
                        probabilities[j] = SmoothPayoffLearning.probs(diffs);
                        probabilities[expectation.length-1] -= probabilities[j];
                }
                
                //Figure out which action to take on this round
                double number = i.getRandom().nextDouble();
                double count = probabilities[0];
                for(int j=0;j<expectation.length;j++) { 
                        if(number <= count)
                                return j;
                        else
                                count += probabilities[j+1];
                }
                
                //Shouldn't ever get here
                return -1;

        }

        public SmoothPayoffLearning(Game g, double gam) throws ModelError{ game = g; gamma = gam; }
}
