package models;
import utilities.Individual;
import utilities.MersenneTwister;
import utilities.ModelError;
import learningRules.SmoothReinforcement;

import com.martiansoftware.jsap.*;

public class SignalingGame {
        
        static final MersenneTwister random = new MersenneTwister();
        
        public static final int SENDER = 0;
        public static final int RECEIVER = 1;
        
        Individual[] player;
        
        int states;
        int signals;
        
        SmoothReinforcement lr;
        
        public void clear() {
                player[SENDER].clearStratWeights(signals, states);
                player[RECEIVER].clearStratWeights(states, signals);
        }
        
        public void run() throws ModelError{
                
                int nowstate = random.nextInt(states);
                player[SENDER].setStrat(lr.pickStrat(player[SENDER], nowstate));
                player[RECEIVER].setStrat(lr.pickStrat(player[RECEIVER], player[SENDER].getStrat()));
//                System.out.print(nowstate + " " + player[SENDER].getStrat() + " " + player[RECEIVER].getStrat());
                if(player[RECEIVER].getStrat() == nowstate) {
                        player[SENDER].setPayoff(2);
                        player[RECEIVER].setPayoff(2);
                        //System.out.print("+");
                }
                else {
                        player[SENDER].setPayoff(0);
                        player[RECEIVER].setPayoff(0);
                       // System.out.print("-");
                }
                //System.out.println(player[SENDER].getWeights(nowstate, player[SENDER].getStrat()) + " " + 
                //        player[RECEIVER].getWeights(player[SENDER].getStrat(), player[RECEIVER].getStrat()));
                lr.processPayoff(player[SENDER], nowstate);
                lr.processPayoff(player[RECEIVER], player[SENDER].getStrat());
                //System.out.println(player[SENDER].getWeights(nowstate, player[SENDER].getStrat()) + " " + 
                //        player[RECEIVER].getWeights(player[SENDER].getStrat(), player[RECEIVER].getStrat()));
        }
        
        /**
        * Detects if a pair is playing a signaling system.  A returns true if the a signaling
        * system is played with probability 0.99 or greater.  Otherwise false.  This does not
        * distinguish between failed learning and learning a non-signaling system.
        */
        
        public boolean detectSigSystem() {

                double[][] sendProbs = new double[states][];
                double[][] recevProbs = new double[signals][];
                double[] weights;
                double[] diffs;
                
                for(int i=0;i<states;i++) {
                        //Calculate probabilities for the sender
                        weights = player[SENDER].getDomainWeights(i);
                        sendProbs[i] = new double[weights.length];
                        
                        //Save calculating one probability
                        sendProbs[i][weights.length-1] = 1.0;
                        diffs = new double[weights.length];
                        for(int j=0;j<(weights.length -1);j++) {
                                for(int k=0;k<weights.length;k++) {
                                        diffs[k] = (double) lr.getBeta() * ((double) weights[k] - weights[j]);
                                }
                                sendProbs[i][j] = SmoothReinforcement.probs(diffs);
                                sendProbs[i][weights.length-1] -= sendProbs[i][j];
                        }
                }
                for(int i=0;i<signals;i++) {
                        //Calculate probabilities for the sender
                        weights = player[RECEIVER].getDomainWeights(i);
                        recevProbs[i] = new double[weights.length];
                        
                        //Save calculating one probability
                        recevProbs[i][weights.length-1] = 1.0;
                        diffs = new double[weights.length];
                        for(int j=0;j<(weights.length -1);j++) {
                                for(int k=0;k<weights.length;k++) {
                                        diffs[k] = (double) lr.getBeta() * ((double) weights[k] - weights[j]);
                                }
                                recevProbs[i][j] = SmoothReinforcement.probs(diffs);
                                recevProbs[i][weights.length-1] -= recevProbs[i][j];
                        }
                }       
                
                boolean sigSys = true;
                for(int i=0;i<states;i++) {
                        boolean goodstate = false;
                        for(int j=0;j<signals;j++) {
                                if(sendProbs[i][j] > 0.99) {
                                        if(recevProbs[j][i] > 0.99)
                                                goodstate = true;
                                }
                        }
                        if(!goodstate) {
                                sigSys = false;
                                break;
                        }
                }
                return sigSys;
        }
        
        public double calculateEfficiency() {
            double[][] sendProbs = new double[states][];
            double[][] recevProbs = new double[signals][];
            double[] weights;
            double[] diffs;
            
            for(int i=0;i<states;i++) {
                    //Calculate probabilities for the sender
                    weights = player[SENDER].getDomainWeights(i);
                    sendProbs[i] = new double[weights.length];
                    
                    //Save calculating one probability
                    sendProbs[i][weights.length-1] = 1.0;
                    diffs = new double[weights.length];
                    for(int j=0;j<(weights.length -1);j++) {
                            for(int k=0;k<weights.length;k++) {
                                    diffs[k] = (double) lr.getBeta() * ((double) weights[k] - weights[j]);
                            }
                            sendProbs[i][j] = SmoothReinforcement.probs(diffs);
                            sendProbs[i][weights.length-1] -= sendProbs[i][j];
                    }
            }
            for(int i=0;i<signals;i++) {
                    //Calculate probabilities for the receiver
                    weights = player[RECEIVER].getDomainWeights(i);
                    recevProbs[i] = new double[weights.length];
                    
                    //Save calculating one probability
                    recevProbs[i][weights.length-1] = 1.0;
                    diffs = new double[weights.length];
                    for(int j=0;j<(weights.length -1);j++) {
                            for(int k=0;k<weights.length;k++) {
                                    diffs[k] = (double) lr.getBeta() * ((double) weights[k] - weights[j]);
                            }
                            recevProbs[i][j] = SmoothReinforcement.probs(diffs);
                            recevProbs[i][weights.length-1] -= recevProbs[i][j];
                    }
            }       
            
        	double success = 0.0;
            for(int i=0;i<states;i++) {
            	for(int j=0;j<signals;j++) {
            		success += ((double) 1/states) * sendProbs[i][j]*recevProbs[j][i];
            	}
            }
            
            return success;
            
        }
        
        /**
        * Outputs the probabilities that each strategy is played to StdOut
        */
        public void printResults() {
                System.out.println("Sender");
                for(int i=0;i<states;i++) {
                        double[] weights = player[SENDER].getDomainWeights(i);
                        
                
                        double[] probabilities = new double[weights.length];
                
                        //Save calculating one probability
                        probabilities[weights.length-1] = 1.0;
                        double[] diffs = new double[weights.length];
                        for(int j=0;j<(weights.length -1);j++) {
                                for(int k=0;k<weights.length;k++) {
                                        diffs[k] = (double) lr.getBeta() * ((double) weights[k] - weights[j]);
                                }
                                probabilities[j] = SmoothReinforcement.probs(diffs);
                                probabilities[weights.length-1] -= probabilities[j];
                        }
                        System.out.println("\tState " + i);
                        for (int j=0;j<signals;j++) {
                                System.out.println("\t\tSignal " + j + ": " + 
                                        probabilities[j] + " (" + weights[j] + ")");
                        }
                }
                System.out.println("Receiver");
                for(int i=0;i<signals;i++) {
                        double[] weights = player[RECEIVER].getDomainWeights(i);
                
                        double[] probabilities = new double[weights.length];
                
                        //Save calculating one probability
                        probabilities[weights.length-1] = 1.0;
                        double[] diffs = new double[weights.length];
                        for(int j=0;j<(weights.length -1);j++) {
                                for(int k=0;k<weights.length;k++) {
                                        diffs[k] = (double) lr.getBeta() * ((double) weights[k] - weights[j]);
                                }
                                probabilities[j] = SmoothReinforcement.probs(diffs);
                                probabilities[weights.length-1] -= probabilities[j];
                        }
                        System.out.println("\tSignal " + i);
                        for (int j=0;j<states;j++) {
                                System.out.println("\t\tAct " + j + ": " + 
                                        probabilities[j] + " (" + weights[j] + ")");
                        }
                }
        }
        
        /**
        * Constructor. 
        * @param t Number of state
        * @param i Number of signals
        * @param d Delta (part of learning rule)
        * @param b Beta (part of learning rule)
        */
        
        public SignalingGame(int t, int i, double d, double b) { 
                states = t; signals = i; 
                lr = new SmoothReinforcement(d, b);
                player = new Individual[2];
                player[SENDER] = new Individual();
                player[SENDER].setLearningRule(lr);
                player[SENDER].setRandom(random);
                player[RECEIVER] = new Individual();
                player[RECEIVER].setLearningRule(lr);
                player[RECEIVER].setRandom(random);
        }
        
        public static final void main(String[] args) {
                try {
                // We're using the JSAP system for argument parsing
                
                JSAP jsap = new JSAP();
                
                FlaggedOption delta = new FlaggedOption("delta")
                                                .setShortFlag('d')
                                                .setLongFlag("delta")
                                                .setRequired(true)
                                                .setDefault("0.99")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                delta.setHelp("Chooses a delta");
                jsap.registerParameter(delta);
                
                FlaggedOption beta = new FlaggedOption("beta")
                                                .setShortFlag('b')
                                                .setLongFlag("beta")
                                                .setRequired(true)
                                                .setDefault("0.5")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                beta.setHelp("Chooses a beta");
                jsap.registerParameter(beta);
                
                FlaggedOption states = new FlaggedOption("states")
                                                .setShortFlag('t')
                                                .setLongFlag("states")
                                                .setRequired(true)
                                                .setDefault("3")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                states.setHelp("Number of states");
                jsap.registerParameter(states);
                
                FlaggedOption sigs = new FlaggedOption("sigs")
                                                .setShortFlag('s')
                                                .setLongFlag("signals")
                                                .setRequired(true)
                                                .setDefault("3")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                sigs.setHelp("Number of signals");
                jsap.registerParameter(sigs);
               
                FlaggedOption gens = new FlaggedOption("gens")
                                                .setShortFlag('g')
                                                .setLongFlag("gens")
                                                .setRequired(true)
                                                .setDefault("10000")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                gens.setHelp("Number of generations on each run");
                jsap.registerParameter(gens);
                
                FlaggedOption runs = new FlaggedOption("runs")
                                                .setShortFlag('r')
                                                .setLongFlag("runs")
                                                .setRequired(true)
                                                .setDefault("1")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                runs.setHelp("Number of runs");
                jsap.registerParameter(runs);
                
                FlaggedOption verbose = new FlaggedOption("verbose")
                                                .setShortFlag('v')
                                                .setLongFlag("verbosity")
                                                .setRequired(true)
                                                .setDefault("0")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                verbose.setHelp("Verbosity, 0 for summary, 1 for run-by-run, 2 for exceptional detail");
                jsap.registerParameter(verbose);
                
                // Process commandline 
                JSAPResult config = jsap.parse(args);   
                if (!config.success()) {
                        System.err.println();
                        System.err.println("Usage: java "
                                + SignalingGame.class.getName());
                        System.err.println("                "
                                + jsap.getUsage());
                        System.err.println();
                        System.err.println(jsap.getHelp());
                        System.exit(1);
                }
                
                SignalingGame game = new SignalingGame(config.getInt("states"),
                                                       config.getInt("sigs"),
                                                       config.getDouble("delta"),
                                                       config.getDouble("beta")); 
                
               int convergeCount=0;
               for(int i=0;i<config.getInt("runs");i++) {
                        if(config.getInt("verbose") == 2)
                               System.out.println("------ New Run ------");
                        game.clear();
                        for(int j=0;j<config.getInt("gens");j++) {
                                game.run();
                        }
                        if(game.detectSigSystem())
                                convergeCount++;
                        if(config.getInt("verbose") == 1)
                        	System.out.println(config.getInt("states") + " " + 
                       				config.getInt("sigs") + " " +
                       				config.getDouble("delta") + " " +
                       				config.getDouble("beta") + " " +
                       				game.calculateEfficiency());
                        if(config.getInt("verbose") == 2)
                                game.printResults();
               }
               if(config.getInt("verbose") == 0) 
            	   System.out.println("------ Summary ------");
               System.out.println(config.getInt("states") + " " + 
                       				config.getInt("sigs") + " " +
                       				config.getDouble("delta") + " " +
                       				config.getDouble("beta") + " " +
                       				((double) convergeCount/config.getInt("runs")));

        }
        catch(ModelError e) {
                System.out.println(e);
        }
        catch(JSAPException e) {
                System.out.println(e);
        }
        }       
}
