package models;

import java.util.*;
import java.io.*;

import utilities.Game;
import utilities.GraphIO;
import utilities.Individual;
import utilities.Jobs;
import utilities.MersenneTwister;
import utilities.ModelError;

import learningRules.BgLearning;
import learningRules.LearningRule;

import com.martiansoftware.jsap.*;

import connectionRules.ConnectNothing;
import connectionRules.ConnectionRule;

/**
* This is a class that runs the Standard Model of Bayesian Learning in Networks.
* 
* @author Kevin Zollman
*
*/


public class StandardModel {

        Game[] games;
        
        /**
        * Sets an array of games, one for each state of the world.
        * @param g The array of Game objects
        */
        public synchronized void setGames(Game[] g) { games = g;}
        
        /**
        * Returns the games for StandardModel
        * @return the games
        */
        public Game[] getGames() { return games; }
        
        // The folks in the network
        Individual[] individuals;
  
        /**
        * Sets the list of individuals for the model.
        * @param i An array of individuals
        */
        public synchronized void setIndividuals(Individual[] i) { individuals = i; }
        
        /**
        * Returns the set of individuals for the model.
        * @return the individuals
        */
        public Individual[] getIndividuals() {return individuals;}

        // The random number generator
        MersenneTwister random;
        
        /**
        * Sets the random number generator.
        * @param r the mersennetwister random number generator
        */
        public synchronized void setRandom(MersenneTwister r) { random =r; }
        
        /** 
        * Returns the random number generator for the model.
        * @return the mersenne twister random number generator
        */
        public MersenneTwister getRandom() { return random; }        
        
 
        
        /**
        * Forms a new network of specified size; doesn't expect much to be set.
        * This method only tells each individual of the network, does not set anything else.
        * @param inds A list of individuals
        */
        public synchronized void createPlayersLow(Individual[] inds) {
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                }
        }
        
        /**
        * Forms a new network of specified size; expects random to be set.
        * This method tells each individual of the network, sets the random number generator,
        * and calle each individuals initStart function.
        *  
        * @param inds A list of individuals
        * @param s The number of states of the world
        * @throws ModelError If an initialization error
        */
        public synchronized void createPlayersHigh(Individual[] inds, int s) throws ModelError {
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                        individuals[i].setRandom(random);
                        individuals[i].initStart(s);
                        
                }
        }
        
        /**
        * Forms a new network of specified size; expects random to be set.
        * This method tells each individual of the network, sets the random number generator,
        * gives the individuals a connection and learning rule, sets the mutation rate, sets 
        * belief constraint, and calls each individuals initStart function
        *
        * @param inds A list of individuals
        * @param s The number of states of the world
        * @param c The connection rule used for forming network connections by each individual
        * @param l The learning rule used for determining strategies by each individual
        * @param m The mutation rate for each individual to reset their beliefs
        * @param z The zero-state constraint for beliefs (initial beliefs cannot be lower than z)
        * @throws ModelError If something fails during initialization
        */
        public synchronized void createPlayersHigh(Individual[] inds, int s, ConnectionRule c, LearningRule l, double m, double z) throws ModelError{
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                        individuals[i].setRandom(random);
                        individuals[i].setConnectionRule(c);
                        individuals[i].setLearningRule(l);
                        individuals[i].setMutation(m);
                        individuals[i].setConstraint(z);
                        individuals[i].initStart(s);

                }
        }
        
        /**
        * Forms a new network of specified size; expects random to be set.
        * This method tells each individual of the network, sets the random number generator,
        * gives the individuals a connection and learning rule, sets the list of games sets the 
        * mutation rate, sets belief constraint, and calls each individuals initStart function
        *
        * @param inds A list of individuals
        * @param s The number of states of the world
        * @param c The connection rule used for forming network connections by each individual
        * @param l The learning rule used for determining strategies by each individual
        * @param g The games used in the model
        * @param m The mutation rate for each individual to reset their beliefs
        * @param z The zero-state constraint for beliefs (initial beliefs cannot be lower than z)
        * @throws ModelError If something fails during initialization
        */
        public synchronized void createPlayersHigh(Individual[] inds, int s, ConnectionRule c, LearningRule l, Game g, double m, double z) throws ModelError{
                individuals = inds;
                for(int i=0;i<inds.length;i++) {
                        individuals[i].setNetwork(individuals);
                        individuals[i].setRandom(random);
                        individuals[i].setConnectionRule(c);
                        individuals[i].setLearningRule(l);
                        individuals[i].setGame(g);
                        individuals[i].setMutation(m);
                        individuals[i].setConstraint(z);
                        individuals[i].initStart(s);
                }
        }
        
        
        /**
        * Plays a generation and then resets.
        * This first calls each individuals reset function, then calls each individuals playGame
        * function, then calles each individuals update function
        * @throws ModelError If something bad happens
        */
        public synchronized void playGen() throws ModelError {
                for(int i=0;i<individuals.length;i++) 
                        individuals[i].reset();
                for(int i=0;i<individuals.length;i++) 
                        individuals[i].playGame();
                for(int i=0;i<individuals.length;i++)
                        individuals[i].update();
        }
        
        // Returns three values, 0=unconverged, 1=converged to the true state, 2=converged to the
        //   wrong state.
        
        /**
        * Checks to see if individuals have converged.  Convergence means beliefs in > 0.9999 in
        * any state.
        * This function presumes that state 0 is the uninformative state.
        * @param sow What is the actual state of the world?
        * @return This function returns 0 if unconverged, 1 if converged to the true state, 
        *         2 if converged to the wrong state.
        */
        
        public synchronized int detectConvergence(int sow) throws ModelError {
                boolean trueconv = true;
                boolean badconv = true;
                boolean badindividconv[] = new boolean[getIndividuals().length];
                for(int k=0;k<getIndividuals().length;k++) {
                        if(getIndividuals()[k].getBelief(sow) < 0.9999) {
                                trueconv = false;
                        }
                        else {
                                badindividconv[k] = false;
                                for(int i=0;i<games.length;i++) {
                                        if((i!=sow)&&(getIndividuals()[k].getBelief(i) >= 0.9999)) {
                                                badindividconv[k]=true;
                                                break;
                                        }
                                }
                        }
                }
                for(int k=0;k<getIndividuals().length;k++) {
                        if(badindividconv[k] == false) {
                                badconv=false;
                                break;
                        }
                }
                if(badconv == true) 
                        return 2;
                if(trueconv == true)
                        return 1;
                else
                        return 0;
        }
        
        
        /**
        * Does nothing
        */
        public StandardModel() {};
        
        /**
        * Creates a new StandardModel with a specified number of individuals.
        * This creates a new list of individuals, sets their id number, and tells each about the 
        * network.
        * @param n The size of the network
        */
        public StandardModel(int n) {
                individuals = new Individual[n];
                for(int i=0;i<n;i++) {
                        individuals[i] = new Individual();
                        individuals[i].setId(n);
                        individuals[i].setNetwork(individuals);
                }
        }
        
        
        /** 
        * Main, runs a model.
        * @param args The command line arguments
        */
        public static void main(String[] args) {
                
                
                try {

                // We're using the JSAP system for argument parsing
                
                JSAP jsap = new JSAP();
                
                // -j <jobfile> is the job-file
                FlaggedOption jobfile = new FlaggedOption("jobFile")
                                                .setShortFlag('j')
                                                .setLongFlag("jobfile")
                                                .setRequired(true)
                                                .setStringParser(JSAP.STRING_PARSER);
                jobfile.setHelp("The job file containing network information");
                jsap.registerParameter(jobfile);
                
                // -v prints detailed output
                Switch switchver = new Switch("verbose")
                                        .setShortFlag('v')
                                        .setLongFlag("verbose");
                switchver.setHelp("Prints information on each run rather than summary statistics");
                jsap.registerParameter(switchver);
                
                // -r <runs> s the number of runs per job, default 40,000
                FlaggedOption runsoption = new FlaggedOption("runs")
                                                .setShortFlag('r')
                                                .setLongFlag("runs")
                                                .setRequired(true)
                                                .setDefault("40000")
                                                .setStringParser(JSAP.LONG_PARSER);
                runsoption.setHelp("Number of runs per job");
                jsap.registerParameter(runsoption);
                
                // -b <payoff> for the bad act, default 50
                FlaggedOption bad = new FlaggedOption("badPayoff")
                                                .setShortFlag('b')
                                                .setLongFlag("badpayoff")
                                                .setRequired(true)
                                                .setDefault("50")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                bad.setHelp("Set the payoff for the bad act");
                jsap.registerParameter(bad);
                
                // -g <generations> is the maximum number of generations until give up
                //                  default 2000
                FlaggedOption generations = new FlaggedOption("generations")
                                                .setShortFlag('g')
                                                .setLongFlag("generations")
                                                .setRequired(true)
                                                .setDefault("2000")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                generations.setHelp("Number of generations per run");
                jsap.registerParameter(generations);
                
                
                // -s <stratnum> gives the number of strategies default 2
                FlaggedOption stratnum = new FlaggedOption("stratnum")
                                                .setShortFlag('s')
                                                .setLongFlag("stratnum")
                                                .setRequired(true)
                                                .setDefault("2")
                                                .setStringParser(JSAP.INTEGER_PARSER);
                stratnum.setHelp("Number of states/strategies");
                jsap.registerParameter(stratnum);
                                
                // -m <rate> sets tbe mutation rate, default 0.0
                FlaggedOption mutationrate = new FlaggedOption("mutationRate")
                                                .setShortFlag('m')
                                                .setLongFlag("mutation")
                                                .setRequired(true)
                                                .setDefault("0.0")
                                                .setStringParser(JSAP.DOUBLE_PARSER);
                mutationrate.setHelp("Sets the probability that an individual randomly changes her beleifs");
                jsap.registerParameter(mutationrate);
                
                // Process commandline 
                JSAPResult config = jsap.parse(args);   
                if (!config.success()) {
                        System.err.println();
                        System.err.println("Usage: java "
                                + BgModel.class.getName());
                        System.err.println("                "
                                + jsap.getUsage());
                        System.err.println();
                        System.err.println(jsap.getHelp());
                        System.exit(1);
                }
                                
                //    Jobs file
                HashSet jobs = GraphIO.readJobsFile(config.getString("jobFile"));
                Iterator todo = jobs.iterator();
                
                //Other initializions
                StandardModel model = new StandardModel();
                model.setRandom(new MersenneTwister());
                
                //Parameters of the model
                boolean stochastic = true;
                ConnectionRule c = new ConnectNothing();
                
                // The second argument gives the number of strategies, we will now generate a
                //   a game for that number of strategies.  If an agent assigns equal probability
                //   to all states, she is indifferent between the actions.
                
                int[][][] matrix = new int[config.getInt("stratnum")][config.getInt("stratnum")][];
                
                //Set up the bad payoff and good payoffs
                int badpayoff = config.getInt("badPayoff");
                int goodpayoff = badpayoff+1;
                
                int[] badpaymatrix = new int[config.getInt("stratnum")];
                int[] goodpaymatrix = new int[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        badpaymatrix[i] = badpayoff;
                        goodpaymatrix[i] = goodpayoff;
                }
                int[][][] matrices = new int[config.getInt("stratnum")][config.getInt("stratnum")][];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        for(int j=0;j<config.getInt("stratnum");j++) {
                                if(j==i)
                                        matrices[i][j] = goodpaymatrix;
                                else 
                                        matrices[i][j] = badpaymatrix;
                        }
                }
                Game games[] = new Game[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) {
                        //creates a new game with the appropriate matrix which is stochastic and
                        //   non-strategic
                        games[i] = new Game(matrices[i], stochastic, model.getRandom(), false);
                }
                
                model.setGames(games);
                
                
                // One more bit of initialization
                BgLearning l = new BgLearning(model.getGames());
                
                System.out.println("Standard Model parameter list: ");
                System.out.println("   Jobfile: " + config.getString("jobFile"));
                System.out.println("   Number of Strategies: " + config.getInt("stratnum"));
                System.out.println("   Verbose: " + config.getBoolean("verbose"));
                System.out.println("   Runs: " + config.getLong("runs"));
                System.out.println("   Bad Payoff: " + config.getInt("badPayoff"));
                System.out.println("   Max. Generations/Run: " + config.getInt("generations"));
                System.out.println("   Mutation Rate: " + config.getDouble("mutationRate"));
                
                while(todo.hasNext()) {
                
                Jobs job = (Jobs) todo.next();
                model.createPlayersHigh(GraphIO.adjacencyMatrix(job.getMatrix()), matrix.length, c, l, config.getDouble("mutationRate"), 0.0);
                
                if(!config.getBoolean("verbose")) {
                        System.out.println("----");
                        System.out.println("Job Id: " + job.getJobID());
                        System.out.println("Job Description: " + job.getDescrip());
                }
                
                long avegen = 0L;
                long goodconverged = 0L;
                long badconverged = 0L;
                long[] statecount = new long[config.getInt("stratnum")];
                for(int i=0;i<config.getInt("stratnum");i++) 
                        statecount[i] = 0L;
                for(long runs=0L;runs<config.getLong("runs");runs++) {
                        int sow = model.getRandom().nextInt(config.getInt("stratnum"));
                        statecount[sow]++;
                        for(int j=0;j<model.getIndividuals().length;j++) {
                                model.getIndividuals()[j].randBeliefs(config.getInt("stratnum"));
                                model.getIndividuals()[j].setGame(model.getGames()[sow]);
                        }
                        int i;
                        int end=0;
                        for(i=0;i<config.getInt("generations");i++) {
                                model.playGen();
                                end = model.detectConvergence(sow);
                                // zero means not converged
                                if(end == 0) 
                                        continue;
                                // one means converged to correct state
                                if(end == 1)
                                        break;
                                // two means converged wrong state
                                if(end == 2)
                                        break;
                        }
                        if(i != config.getInt("generations")-1) {
                                if(end == 2) {
                                        badconverged++;
                                }
                                else {
                                        goodconverged++;
                                        avegen += i;
                                }
                        }
                        if(config.getBoolean("verbose")) {
                                System.out.println(job.getJobID() + " " + job.getDescrip() + 
                                        " " + runs + " " + sow + " " + end + " " + i);
                        }
                        else {
                                if(runs%250==0) 
                                        System.out.print("*");
                        }
                }
                if(!config.getBoolean("verbose")) {
                        System.out.println();
                        System.out.println(badconverged + " runs incorrectectly converged");
                        System.out.println("  Correct convergence took an average of " +
                                (double) avegen/goodconverged + " generations.  ");
                }        
        }
        }
        catch(ModelError e) {
                System.out.println(e);
        }
        catch(FileNotFoundException e) {
                System.out.println(e);
        }
        catch(IOException e) {
                System.out.println(e);
        }
        catch(JSAPException e) {
                System.out.println(e);
        }
        }
}
