/**
 * 
 */
package mutationRules;

import utilities.Individual;
import utilities.ModelError;

/**
 * This mutates strategies based on a mutation matrix
 * @author Kevin J.S. Zollman
 *
 */
public class MutateStrategy implements MutationRule {

	double[][] mutMatrix;
	
	/** 
	 * This function mutates based on the mutation matrix
	 * already established by the constructor.  It changes
	 * Individual.strategy without setting LastStrategy
	 * @param i The Calling Individual
	 */
	public void mutate(Individual i) throws ModelError  {
		double n = i.getRandom().nextDouble();
		double total = 0.0;
		synchronized(i) {
		for(int j=0;j<mutMatrix[i.getStrat()].length;j++) {
			total += mutMatrix[i.getStrat()][j];
			if(n < total) {
				i.setStrat(j);
				break;
			}
		}
		}
	}
	
	/**
	 * Sets the mutation rate (equivalent to the constructor)
	 */
	
	public void setMutationRate(double m, int s) throws ModelError {
		if((m > 1.0)||(m<0.0)) {
			throw new ModelError("M not a probability:" + m);
		}
		
		mutMatrix = new double[s][s];
		
		double im = (double) m/(s-1);
		for(int i=0;i<s;i++) {
			for(int j=0;j<s;j++) {
				if(i==j) {
					mutMatrix[i][j] = (double) 1-m;
				}
				else {
					mutMatrix[i][j] = im;
				}
			}
		}
	}

	/**
	 * Same as constructor
	 * @param m The mutation matrix where [i][j] is the 
	 *          probability that strategy i mutates to 
	 *          strategy j.
	 * @throws ModelError
	 */
	public void setMutationRate(double[][] m) throws ModelError { 
		mutMatrix = m;
		if(mutMatrix.length != mutMatrix[0].length) {
			throw new ModelError("Mutation Matrix not a square matrix");
		}
		for(int i=0;i<mutMatrix.length;i++) {
			double total = 0.0; 
			for(int j=0;j<mutMatrix[i].length;j++) {
				total += mutMatrix[i][j];
			}
			if(total != 1.0) {
				throw new ModelError("Row " + i + " does not sum to 1.0");
			}
		}
	}

	/**
	 * Because there is no single mutation rate this
	 * is undefined.
	 */
	public double getMutationRate() { return -1.0; }
	
	/**
	 * The constructor which takes a mutation matrix
	 * @param m The mutation matrix where [i][j] is the 
	 *          probability that strategy i mutates to 
	 *          strategy j.
	 * @throws ModelError
	 */
	public MutateStrategy(double[][] m) throws ModelError { 
		mutMatrix = m;
		if(mutMatrix.length != mutMatrix[0].length) {
			throw new ModelError("Mutation Matrix not a square matrix");
		}
		for(int i=0;i<mutMatrix.length;i++) {
			double total = 0.0; 
			for(int j=0;j<mutMatrix[i].length;j++) {
				total += mutMatrix[i][j];
			}
			if(total != 1.0) {
				throw new ModelError("Row " + i + " does not sum to 1.0");
			}
		}
	}
	
	/**
	 * Sets the mutation matrix as uniform based on m
	 * @param m The mutation rate (the probability that
	 * 			a given strategy switches to any other strategy)
	 * 			
	 * @param s The number of strategies
	 * @throws ModelError
	 */
	public MutateStrategy(double m, int s) throws ModelError {
		if((m > 1.0)||(m<0.0)) {
			throw new ModelError("M not a probability:" + m);
		}
		
		mutMatrix = new double[s][s];
		
		double im = (double) m/(s-1);
		for(int i=0;i<s;i++) {
			for(int j=0;j<s;j++) {
				if(i==j) {
					mutMatrix[i][j] = (double) 1-m;
				}
				else {
					mutMatrix[i][j] = im;
				}
			}
		}
	}
}
