package utilities;
import java.math.*;

/**
* A class to compute a binomial function.  There are two options with this class.  First,
*  you might just call longBinom as a static function, which will calculate the binomial
*  you seek on the spot.  If you will be doing alot of similar calculations a better option
*  is to initialize the function with a largest N, this stores all the binomials in memory
*  which (although big) has the advantage of saving <i>a lot</i> of time.  Since it takes up
*  so much memory, it is worth getting rid of as quickly as possible (i.e. make it with very
*  small scope)!
* @author Kevin Zollman
*/
public class Binomial {

    BigDecimal[][] binomials;
    int globalN;
    
    /**
    * Determines the binomial function.  This does everything locally, it is used if
    *   a binomial is called without being initialized first.
    * @param N n of the binomial function
    * @param k k of the binomial function
    * @return n choose k
    */
    public static BigDecimal longBinom(int N, int K) {

    	BigDecimal one = new BigDecimal(1.0);
        BigDecimal zero = new BigDecimal(0.0);
        
        BigDecimal[][] binoms = new BigDecimal[N+1][];
        for(int i=0;i<N+1;i++) binoms[i] = new BigDecimal[i+2];

        binoms[0][1] = zero;
        for (int n = 0; n <= N; n++) binoms[n][0] = one;

        // This is tedious, I know, but it reduces the memory footprint
        // to about 25% of what it would be otherwise.
        for (int n = 1; n <= N; n++) {
            for (int k = 1; k <= n+1; k++) {
            	if(k==n) {
            		binoms[n][k] = one;
            	}
            	else {
            		if(k>n) {
            			binoms[n][k] = zero;
            		}
            		else {
            			if(k>(n/2)) {
            				binoms[n][k] = binoms[n][n-k];
            			}
            			else {
            				binoms[n][k] = binoms[n-1][k-1].add(binoms[n-1][k]);
            			}
            		}
            	}
            }
        }

        return binoms[N][K];
    }
    
    /**
    * Determines a binomial function N choose K.  It checks to see if the result is in the
    *   calculated range first, if it is this function returns the value.  Otherwise it 
    *   calls longBinom() and calculates it on the spot.
    * @param N n
    * @param K k
    * @return N choose K
    */
    
    public BigDecimal binom(int N, int K) {
    		if(K>N) 
    				return new BigDecimal(0.0);
    		if((globalN > 0) && (N<=globalN)) 
                    return binomials[N][K];
            else 
                    return longBinom(N, K);
    }
    
    /**
    * Constructor that does nothing
    */
    public Binomial() { globalN = -1; }
    
    /**
    * Constructor that builds a class for speedy multiple callse
    * @param N Biggest N for future use
    */
    public Binomial(int N) {
        BigDecimal one = new BigDecimal(1.0);
        BigDecimal zero = new BigDecimal(0.0);
        
    	binomials = new BigDecimal[N+1][];
        for(int i=0;i<N+1;i++) binomials[i] = new BigDecimal[i+2];

        globalN = N;
        binomials[0][1] = zero;
        for (int n = 0; n <= N; n++) binomials[n][0] = one;

        // This is tedious, I know, but it reduces the memory footprint
        // to about 25% of what it would be otherwise.
        for (int n = 1; n <= N; n++) {
            for (int k = 1; k <= n+1; k++) {
            	if(k==n) {
            		binomials[n][k] = one;
            	}
            	else {
            		if(k>n) {
            			binomials[n][k] = zero;
            		}
            		else {
            			if(k>(n/2)) {
            				binomials[n][k] = binomials[n][n-k];
            			}
            			else {
            				binomials[n][k] = binomials[n-1][k-1].add(binomials[n-1][k]);
            			}
            		}
            	}
            }
        }
    }
}	
