package utilities;
import java.math.*;
/**
* Holds a standard two player game.  A game has several options.  Strategic: not used by this
* class, but says whether the payoff depends on other players (this is not independently
* calculated by the class, you must set it).  Stochastic: if the payoffs provided represent the 
* mean of the payoffs returned or if the payoffs returned should be exactly the one in the matrix.
*
* @author Kevin Zollman
*/
public class Game {
        boolean strategic;
        /**
        * A function to set if the game is strategic.  A game is strategic if the payoffs 
        * are determined by the other actors.  This is not used in the class at all, but is used
        * by the individual class in order to decide if the game is played more than once.
        * @param st A bolean to say if the game is strategic or not
        */
        public void setStrategic(boolean st) { strategic = st; }
        /**
        * A function to say if the game is marked as strategic.
        * @return if the game is strategic
        */
        public boolean isStrategic() { return strategic; }
        
        // A parameter indicating if the payoffs are stochastic
        // Currently the only supported type are non-stochastic or an even distribution between
        //    0 and 2n (integer part only)
        
        boolean stochastic;
        /**
        * A function to set if the game is stochastic.  If the game is stochastic the payoff
        * provided is the mean of a binomial distribution.  If the game is not stochatic, the 
        * payoff returned is exactly the one in the matrix.
        * @param s A boolean to say if the game is stochastic or not
        */
        public void setStochastic(boolean s) {stochastic = s;}
        
        /**
        * A function to determine if the game is stochastic.
        * @return if the game is stochastic
        */
        public boolean getStochastic() {return stochastic; }

        // The actual payoff matrix (only symmetric games supported)       
        int matrix[][]; 
        int bigPayoff;
        int highestMatrix;
        double probs[][][];
        
        /**
        * Gets the payoff matrix
        * @return the two dimensional integer matrix that represents the payoffs of the game
        */
        public int[][] getMatrix() { return matrix; }
        
        /**
        * Sets the payoff matrix.  This function presumes that the stochastic value is already set.
        * Matrices cannot have negative payoffs, this would screw up reinforcement learning.  If
        * the game is stochastic, it calculates the highest possible payoff as being the highest 
        * matrix payoff + the smalles matrix payoff.  The distribution is then a binomial 
        * distribution, from 0 to the highest possible payoff with a mean of the value provided
        * in the matrix.
        * @param m A two dimensional integer array with the matrix
        * @throws ModelError if the matrix is malformed
        */
        public synchronized void setMatrix(int[][] m) throws ModelError{
                for(int i=0;i<m.length;i++) {
                        for(int j=0;j<m[i].length;j++) {
                                if(m[i][j] < 0) 
                                        throw new ModelError("Matrix has negative payoff");
                        }
                }
                matrix = m;
                int hp = 0;
                int sp = 2147483646;
                for(int i=0;i<matrix.length;i++) {
                        for(int j=0;j<matrix[i].length;j++) {
                                if(matrix[i][j] > hp) 
                                        hp = matrix[i][j];
                                if(matrix[i][j] < sp)
                                        sp = matrix[i][j];
                        }
                }
                highestMatrix = hp;
                bigPayoff = hp + sp;
                if(stochastic == true) {
                        probs = new double[matrix.length][matrix[0].length][bigPayoff+1];
                        Binomial binoms = new Binomial(bigPayoff);
                        for(int i=0;i<matrix.length;i++) {
                                for(int j=0;j<matrix[i].length;j++) {
                                        for(int k=0;k<=bigPayoff;k++) {
                                             BigDecimal p = new BigDecimal((double) Math.pow((double) matrix[i][j]/bigPayoff,(double) k)*Math.pow((double) 1-((double) matrix[i][j]/bigPayoff),(double) (bigPayoff-k)));
                                             BigDecimal q = p.multiply(binoms.binom(bigPayoff, k));
                                             probs[i][j][k] = q.doubleValue();
                                        }
                                }
                        }
                }
        }
        
        /**
        * Sets the payoff matrix.  
        * Matrices cannot have negative payoffs, this would screw up reinforcement learning.  The
        * distribution of payoffs returned for stochastic games is a binomial distribution between
        * 0 and bp with a mean of the value provided in the matrix
        * @param m The payoff matrix
        * @param bp The bigest possible payoff (for stochastic games).
        * @throws ModelError if there is a problem with the matrix
        */
        public synchronized void setMatrix(int[][] m, int bp) throws ModelError{
                 for(int i=0;i<m.length;i++) {
                        for(int j=0;j<m[i].length;j++) {
                                if(m[i][j] < 0) 
                                        throw new ModelError("Matrix has negative payoff");
                        }
                }
                matrix = m;
                int hp = 0;
                for(int i=0;i<matrix.length;i++) {
                        for(int j=0;j<matrix[i].length;j++) {
                                if(matrix[i][j] > hp) 
                                        hp = matrix[i][j];
                        }
                }
                highestMatrix = hp;
                bigPayoff = bp;
                if(stochastic == true) {
                        probs = new double[matrix.length][matrix[0].length][bigPayoff+1];
                        Binomial binoms = new Binomial(bigPayoff);
                        for(int i=0;i<matrix.length;i++) {
                                for(int j=0;j<matrix[i].length;j++) {
                                        for(int k=0;k<=bigPayoff;k++) {
                                             BigDecimal p = new BigDecimal((double) Math.pow((double) matrix[i][j]/bigPayoff,(double) k)*Math.pow((double) 1-((double) matrix[i][j]/bigPayoff),(double) (bigPayoff-k)));
                                             BigDecimal q = p.multiply(binoms.binom(bigPayoff, k));
                                             probs[i][j][k] = q.doubleValue();                                        }
                                }
                        }
                }
        }
        
        /**
        * Sets the biggest possible payoff, and then recalculates the probabilites based on that.
        * The payoffs returned are binomially distributed between 0 and bp
        * @param bp the big payoff
        */
        public synchronized void setBigPayoff(int bp) {
                bigPayoff = bp;
                if(stochastic == true) {
                        probs = new double[matrix.length][matrix[0].length][bigPayoff+1];
                        Binomial binoms = new Binomial(bigPayoff);
                        for(int i=0;i<matrix.length;i++) {
                                for(int j=0;j<matrix[i].length;j++) {
                                        for(int k=0;k<=bigPayoff;k++) {
                                             BigDecimal p = new BigDecimal((double) Math.pow((double) matrix[i][j]/bigPayoff,(double) k)*Math.pow((double) 1-((double) matrix[i][j]/bigPayoff),(double) (bigPayoff-k)));
                                             BigDecimal q = p.multiply(binoms.binom(bigPayoff, k));
                                             probs[i][j][k] = q.doubleValue();                                        }
                                }
                        }
                }
        }
        
        /**
        * Tells you the number of strategies
        * @return the number of strategies in the game
        */
        public int getStrats() { return matrix[0].length; }
        
        /**
        * Tells you the highest possible value that will be returned.  In stochastic games this
        * is the biggest possible value (different from the larges value in the matrix). In non-
        * stochastic games this is the highest value in the matrix.
        * @return the biggest possible payoff that will be returned
        */
        public int getHighestPayoff() {
                if(stochastic) 
                        return bigPayoff;
                else
                        return highestMatrix;
        }
        
        /**
        * Tells you the probability of receiving a particular payoff given that the player
        * played strategy r against c.
        * @param r the player's strategy
        * @param c the opponent's strategy
        * @param p the payoff received
        * @return the probability of the player getting that payoff given r and c
        */
        public double getProb(int r, int c, int p) { return probs[r][c][p]; }
        
        /**
        * Tells you the probabiliy of receiving a particular payoff given that the player
        * played strategy r against strategy 0.  This should only be used for non-strategic games
        * otherwise it gives you strange results.
        * @param r the player's strategy
        * @param p the payoff received
        * @return the probability of that player getting that payoff given r
        */
        public double getProb(int r, int p) { return probs[r][0][p]; }
        
        // Ye ole random number generator (only needed for stochastic games)
        MersenneTwister random;
        
        /**
        * Returns the random number generator used
        */
        public MersenneTwister getRandom() {return random;}
        
        /**
        * Sets the random number generator
        */
        public synchronized void setRandom(MersenneTwister r) { random = r;}
        
        // Used for non-strategic games, where the opponent's strategy does not matter
        /**
        * Returns the payoff for strategy r against strategy 0, only usefully for non-strategic
        * games.  The payoff returned will be between zero and the highest payoff.  This will be
        * the same for every call if the game is non-stochastic, but different if stochastic.
        * @param r the strategy employed
        * @return the payoff
        */
        public int getPayoff(int r) {
                return getPayoff(r, 0);
        }
        
        /** 
        * Returns the payoff for strategy r against strategy c.  The payoff returned will be between zero and the highest payoff.  This will be
        * the same for every call if the game is non-stochastic, but different if stochastic.
        * @param r the strategy employed
        * @param c the strategy of the opponent
        * @return the payoff
        */
        public int getPayoff(int r, int c) { 
                if(stochastic) {
                       double n = random.nextDouble();
                       for(int i=0;i<=bigPayoff;i++) {

                               if(n<=probs[r][c][i])        
                                       return i;
                               else n -= probs[r][c][i]; 
                       }
                       return bigPayoff;
                }
                else {
                        return matrix[r][c];
                }
        }
        
        /**
        * Returnes a normalized payoff as a double between 0 and 1 of r played against strategy
        * 0.  This should only be used with non-strategic games.  The payoff will be
        * the same for every call if the game is non-stochastic, but different if stochastic.
        * @param r the strategy
        * @return a double in [0,1]
        */
        public double getNormPayoff(int r) {
                return getNormPayoff(r, 0);
        }
        
        /**
        * Returnes a normalized payoff as a double between 0 and 1 of r played against strategy
        * c. The payoff will be
        * the same for every call if the game is non-stochastic, but different if stochastic.
        * @param r the strategy
        * @param c the opponent's strategy
        * @return a double in [0,1]
        */
        public double getNormPayoff(int r, int c) {
                if(stochastic)
                        return (double) getPayoff(r, c)/bigPayoff;
                else
                        return (double) getPayoff(r, c)/highestMatrix;
        }
        
        /**
        * Returns the expected payoff of strategy r against strategy 0.  This function is used
        * to circumvent stochastic games and get what is actually in the supplied matrix.  This
        * function should only be used in non-strategic games
        * @param r the strategy
        * @return the expected payoff
        */
        public int getExpectedPayoff(int r) { return getExpectedPayoff(r,0); }
        
        /**
        * Returns the expected payoff of strategy r against strategy c.  This function is used
        * to circumvent stochastic games and get what is actually in the supplied matrix.
        * @param r the strategy
        * @param c the opponent's strategy
        * @return the expected payoff
        */
        public int getExpectedPayoff(int r, int c) { return matrix[r][c]; }
        
        /**
        * Returns a game object.  Presumes the game is strategic.
        * @param so a boolean to determine if the game is stochastic or not
        */
        public Game(boolean so) {stochastic = so; strategic = true; }
        
        /**
        * Returns a game object.  Presumes the game is strategic.
        * @param m is the matrix of the game
        * @param so is a boolean to determine if the game is stochstic or not
        * @throws ModelError if there is a problem with the matrix
        */
        public Game(int[][] m, boolean so) throws ModelError{ 
                this(so); 
                for(int i=0;i<m.length;i++) {
                        for(int j=0;j<m[i].length;j++) {
                                if(m[i][j] < 0) 
                                        throw new ModelError("Matrix has negative payoff");
                        }
                }
                matrix = m;
                int hp = 0;
                int sp = 2147483646;
                for(int i=0;i<matrix.length;i++) {
                        for(int j=0;j<matrix[i].length;j++) {
                                if(matrix[i][j] > hp) 
                                        hp = matrix[i][j];
                                if(matrix[i][j] < sp)
                                        sp = matrix[i][j];
                        }
                }
                highestMatrix = hp;
                bigPayoff = hp + sp;
                if(so == true) {
                        probs = new double[matrix.length][matrix[0].length][bigPayoff+1];
                        Binomial binoms = new Binomial(bigPayoff);
                        for(int i=0;i<matrix.length;i++) {
                                for(int j=0;j<matrix[i].length;j++) {
                                        for(int k=0;k<=bigPayoff;k++) {
                                             BigDecimal p = new BigDecimal((double) Math.pow((double) matrix[i][j]/bigPayoff,(double) k)*Math.pow((double) 1-((double) matrix[i][j]/bigPayoff),(double) (bigPayoff-k)));
                                             BigDecimal q = p.multiply(binoms.binom(bigPayoff, k));
                                             probs[i][j][k] = q.doubleValue();
                                        }
                                }
                        }
                }
        }
        
        /**
        * Returns a game object.  Presumes the game is strategic
        * @param m the game matrix
        * @param so a boolean to determin if the game is stochastic
        * @param r the random number generator
        * @throws ModelError if there is a problem with the matrix
        */
        public Game(int[][] m, boolean so, MersenneTwister r) throws ModelError {this(m,so); random = r;}
        
        /**
        * Returns a game object
        * @param m the game matrix
        * @param so a boolean to determine if the game is stochastic
        * @param r the random number generator
        * @param st a boolean to determine if the game is strategic
        * @throws ModelError if there is a problem with the matrix
        */
        public Game(int[][] m, boolean so, MersenneTwister r, boolean st) throws ModelError {
                this(m,so, r); strategic = st;}
        /**
        * Returns a game object.  Presumes the game is strategic.
        * @param so a boolean to determine if the game is stochastic
        * @param r the random number generator
        */
        public Game(boolean so, MersenneTwister r) {stochastic = so; random = r; strategic=true;}

}
